# -*- coding: utf-8 -*-

import time
import json

from django.utils import six
from django.db import connections
from django.utils.encoding import force_text


def ready_sql_logger(sql_logger):
    for connection in connections.all():
        set_custom_cursor(connection, sql_logger)


def set_custom_cursor(connection, logger):
    if not hasattr(connection, 'actionlog_cursor'):
        connection.actionlog_cursor = connection.cursor

    def cursor():
        return CustomCursorWrapper(connection.actionlog_cursor(), connection, logger)

    connection.cursor = cursor


class SqlLogger(object):

    def __init__(self):
        self._sql_time = 0
        self._num_queries = 0
        self._queries = []
        self._databases = {}

    @property
    def sql_count(self):
        return self._num_queries

    @property
    def sql_time(self):
        return round((self._sql_time / 1000), 2)

    def record(self, alias, **kwargs):
        self._queries.append((alias, kwargs))
        if alias not in self._databases:
            self._databases[alias] = {
                'time_spent': kwargs['duration'],
                'num_queries': 1,
            }
        else:
            self._databases[alias]['time_spent'] += kwargs['duration']
            self._databases[alias]['num_queries'] += 1
        self._sql_time += kwargs['duration']
        self._num_queries += 1


class CustomCursorWrapper(object):

    def __init__(self, cursor, db, logger):
        self.cursor = cursor
        self.db = db
        self.logger = logger

    def _quote_expr(self, element):
        if isinstance(element, six.string_types):
            return "'%s'" % force_text(element).replace("'", "''")
        else:
            return repr(element)

    def _quote_params(self, params):
        if not params:
            return params
        if isinstance(params, dict):
            return dict((key, self._quote_expr(value))
                        for key, value in params.items())
        return list(map(self._quote_expr, params))

    def _decode(self, param):
        try:
            return force_text(param, strings_only=True)
        except UnicodeDecodeError:
            return '(encoded string)'

    def _record(self, method, sql, params):
        start_time = time.time()
        try:
            return method(sql, params)
        finally:
            stop_time = time.time()
            duration = (stop_time - start_time) * 1000
            stacktrace = []
            _params = ''
            try:
                _params = json.dumps(list(map(self._decode, params)))
            except Exception:
                pass  # object not JSON serializable

            alias = getattr(self.db, 'alias', 'default')
            conn = self.db.connection
            vendor = getattr(conn, 'vendor', 'unknown')

            params = {
                'vendor': vendor,
                'alias': alias,
                'sql': self.db.ops.last_executed_query(
                    self.cursor, sql, self._quote_params(params)),
                'duration': duration,
                'raw_sql': sql,
                'params': _params,
                'stacktrace': stacktrace,
                'start_time': start_time,
                'stop_time': stop_time,
                'is_select': sql.lower().strip().startswith('select'),
            }
            self.logger.record(**params)

    def execute(self, sql, params=()):
        return self._record(self.cursor.execute, sql, params)

    def __getattr__(self, attr):
        return getattr(self.cursor, attr)

    def __iter__(self):
        return iter(self.cursor)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, print_function

from . import handler_manager
from . import log_format


class Stdout(object):
    """
    Output Standard Out
    """
    def __init__(self, config):
        pass

    def emit(self, messages):
        if not messages.get('is_middleware'):
            output = log_format.dict_to_str(messages)
            print(output)
        else:
            if 'ex_type' in messages:
                output = log_format.error_log(messages)
            else:
                output = log_format.standard_log(messages)
            print(output)

handler_manager.register('stdout', Stdout)



# -*- coding: utf-8 -*-

from . import handler_manager


class Nullout(object):

    def __init__(self, config):
        pass

    def emit(self, data):
        pass

handler_manager.register('null', Nullout)



# coding: utf-8

from fluent import event
from fluent import sender

from . import handler_manager


class Fluentd(object):
    """
    Output Fluentd
    """
    def __init__(self, config):
        host = config.get('host', 'localhost')
        port = config.get('port', 24224)
        tag_name = config.get('tag_name', 'django.actionlog')
        sender.setup(tag_name, host=host, port=port)

    def emit(self, messages):
        is_middleware = messages['is_middleware']
        del messages['is_middleware']

        if is_middleware:
            event.Event('footprint', messages)
        else:
            label_name = ''
            if messages.get('label_name'):
                label_name = messages.get('label_name')
                del messages['label_name']
            event.Event(label_name, messages)

handler_manager.register('fluentd', Fluentd)



# -*- coding: utf-8 -*-

try:
    from django.conf import settings
    ACTION_LOG_SETTING = settings.ACTION_LOG_SETTING
except AttributeError:
    ACTION_LOG_SETTING = {'handler_type': 'null', 'logfile': '/tmp/django_action.log'}

from .handler import handler_manager


class ActionLog(object):

    def __init__(self, config, is_middleware=False):
        self._output = handler_manager.get(config['handler_type'])(config)
        self._is_middleware = is_middleware

    def log(self, **kwargs):
        """
        Output Log
        """
        record = dict(kwargs)
        record.update({'is_middleware': self._is_middleware})
        self._output.emit(record)


def output(messages):
    actlog = ActionLog(ACTION_LOG_SETTING)
    actlog.log(**messages)



# -*- coding: utf-8 -*-

_MODULE_MAP = {}


def register(name, handler_class):
    global _MODULE_MAP
    _MODULE_MAP[name] = handler_class


def get(module_name):
    __import__('{}.{}'.format(_get_package(__name__), module_name))
    return _MODULE_MAP[module_name]


def _get_package(package_path):
    return '.'.join(package_path.split('.')[:-1])



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import datetime


OUTPUT_STANDARD = '================== Footprint Log ==================\n' \
                  'time        : {}\n' \
                  'url         : {}\n' \
                  'view_name   : {}\n' \
                  'request_id  : {}\n' \
                  'status_code : {}\n' \
                  'method      : {}\n' \
                  'user        : {}\n' \
                  'remote_ip   : {}\n' \
                  'user_agnet  : {}\n' \
                  'sql_count   : {}\n' \
                  'sql_time    : {}\n' \
                  'python_time : {}\n' \
                  'total_time  : {}\n\n'


OUTPUT_ERROR = '================== Footprint Log ==================\n' \
               'time        : {}\n' \
               'url         : {}\n' \
               'view_name   : {}\n' \
               'status_code : {}\n' \
               'method      : {}\n' \
               'user        : {}\n' \
               'remote_ip   : {}\n' \
               'user_agnet  : {}\n' \
               'ex_type     : {}\n' \
               'ex_message  : {}\n\n'


def _get_time_str():
    return datetime.datetime.now().strftime(u'%Y/%m/%d %H:%M:%S')


def standard_log(messages):
    output = OUTPUT_STANDARD.format(_get_time_str(),
                                    messages['url'],
                                    messages['view_name'],
                                    messages['request_id'],
                                    messages['status_code'],
                                    messages['method'],
                                    messages['user'],
                                    messages['remote_ip'],
                                    messages['user_agent'],
                                    messages['sql_count'],
                                    messages['sql_time'],
                                    messages['python_time'],
                                    messages['total_time'])
    return output


def error_log(messages):
    output = OUTPUT_ERROR.format(_get_time_str(),
                                 messages['url'],
                                 messages['view_name'],
                                 messages['status_code'],
                                 messages['method'],
                                 messages['user'],
                                 messages['remote_ip'],
                                 messages['user_agent'],
                                 messages['ex_type'],
                                 messages['ex_message'])
    return output


def dict_to_str(messages):
    """
    :param dict messages: {'fizz': 'buzz', ...}
    :rtype: str
    """
    def get_max_length(msg_key_list, max_length=4):
        if msg_key_list:
            max_length = max_length if max_length >= len(msg_key_list[0]) else len(msg_key_list[0])
            return get_max_length(msg_key_list[1:], max_length)
        else:
            return max_length
    del messages['is_middleware']
    max_length = get_max_length(tuple(messages)) + 1
    output = '\n------------------ Action Log ------------------\n'
    output += '{: <{}}: {}\n'.format('time', max_length, _get_time_str())
    for key, val in messages.items():
        output += '{: <{}}: {}\n'.format(key, max_length, val)
    output += '\n'
    return output



# -*- coding: utf-8 -*-

import os
from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

version = __import__('django_actionlog').__version__

setup(
    name='django-actionlog',
    version=version,
    description="You can check the python-time, sql-time and query-count for each request for Django",
    long_description=read('README.rst'),
    author='Ryo Fujimoto',
    author_email='fujimisakari@gmail.com',
    url='http://github.com/fujimisakari/django-actionlog',
    license='BSD',
    packages=['django_actionlog', 'django_actionlog.handler'],
    install_requires=['fluent-logger'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Utilities',
    ],
)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import time

from django.core.urlresolvers import resolve, Resolver404

try:
    from django.conf import settings
    ACTION_LOG_SETTING = settings.ACTION_LOG_SETTING
except AttributeError:
    ACTION_LOG_SETTING = {'handler_type': 'null'}

try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    MiddlewareMixin = object

from .sql_logger import SqlLogger, ready_sql_logger
from .actionlog import ActionLog

_action_log = ActionLog(ACTION_LOG_SETTING, is_middleware=True)
_is_enable = True if ACTION_LOG_SETTING['handler_type'] != 'null' else False


class ActionLogMiddleware(MiddlewareMixin):

    def process_view(self, request, view_func, view_args, view_kwargs):
        if _is_enable:
            request.actionlog_start = time.time()
            request.sql_logger = SqlLogger()
            ready_sql_logger(request.sql_logger)

    def process_response(self, request, response):
        if not hasattr(request, 'actionlog_start'):
            return response

        message = self._create_log_message(request)
        message['status_code'] = response.status_code

        sql_logger = request.sql_logger
        message['sql_count'] = sql_logger.sql_count
        message['sql_time'] = sql_logger.sql_time
        message['python_time'] = round((message['total_time'] - sql_logger.sql_time), 2)
        _action_log.log(**message)
        return response

    def process_exception(self, request, ex):
        if not hasattr(request, 'actionlog_start'):
            return

        message = self._create_log_message(request)
        message['status_code'] = 500
        message['ex_type'] = ex.__class__.__name__
        try:
            message['ex_message'] = unicode(ex.message)
        except:
            message['ex_message'] = str(ex)
        _action_log.log(**message)

    def _create_log_message(self, request):
        path = request.path
        try:
            view_name = resolve(path).view_name
        except Resolver404:
            view_name = '{}'.format(path)

        user_name = ''
        login_obj_name = ACTION_LOG_SETTING.get('login_obj_name', 'user')
        if hasattr(request, login_obj_name):
            login_obj = getattr(request, login_obj_name)
            login_check_func = ACTION_LOG_SETTING.get('login_check_func', 'is_authenticated')
            if getattr(login_obj, login_check_func)():
                if hasattr(login_obj, 'actionlog_name'):
                    user_name = login_obj.actionlog_name
                else:
                    user_name = login_obj.id

        message = {'url': '{}?{}'.format(path, request.GET.urlencode()) if request.GET.urlencode() else path,
                   'request_id': id(request),
                   'view_name': view_name,
                   'method': request.method,
                   'user': user_name,
                   'remote_ip': request.META.get('REMOTE_ADDR', ''),
                   'user_agent': request.META.get('HTTP_USER_AGENT', ''),
                   'total_time': round((time.time() - request.actionlog_start), 2)}
        return message



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from . import handler_manager
from . import log_format


class File(object):
    """
    Output to File
    """
    def __init__(self, config):
        self.logfile = config['logfile'] if config.get('logfile') else '/tmp/django_action.log'

    def emit(self, messages):
        if not messages.get('is_middleware'):
            output = log_format.dict_to_str(messages)
            self.write_to_file(output)
        else:
            if 'ex_type' in messages:
                output = log_format.error_log(messages)
            else:
                output = log_format.standard_log(messages)
            self.write_to_file(output)

    def write_to_file(self, output):
        with open(self.logfile, 'a') as f:
            f.write(output.encode('utf8'))

handler_manager.register('file', File)
